/* File:    Segway_backend.c
 * 
 * Author:  David Zemon
 * Project: Segway
 */

// Includes
#include "Segway_backend.h"

void segway_backend (void *par) {
	int8_t err;
	uint8_t i;
	uint32_t prevTimeCnt = 0;
	int16_t accelFilterValues[ACCEL_ARRAY_SIZE];
	int16_t gyroFilterValues[GYRO_ARRAY_SIZE];
	int16_t *filterValues[NUM_FILTERS] = {accelFilterValues, gyroFilterValues};
	uint8_t accelFilterWeights[ACCEL_ARRAY_SIZE];
	uint8_t gyroFilterWeights[GYRO_ARRAY_SIZE];
	uint8_t *filterWeights[NUM_FILTERS] =
			{accelFilterWeights, gyroFilterWeights};
	filter_t filters[NUM_FILTERS];

	backend_setup(filters, filterValues, filterWeights);

	while (1) {
		// If another thread through an error, stop
		if (g_hardFault)
			while (1)
				;

		// Update Gyroscope LEAN_TO_TURN_READ_RATIO times...
		for (i = 0; i < LEAN_TO_TURN_READ_RATIO; ++i)
			if ((err = readEssentials(&prevTimeCnt, filters)))
				error(err, NULL);

		if ((err = readNonEssentials()))
			error(err, NULL);
	}
}

static void backend_setup (filter_t filters[], int16_t *filterValues[],
		uint8_t *filterWeights[]) {
	int8_t err;
	uint8_t i, j;
	GPIODirModeSet(DEBUG_LED, GPIO_DIR_OUT);
	HD44780Start(LCD_DATA_MASK, LCD_RS, LCD_RW, LCD_EN, LCD_BITMODE, LCD_DIM);
	HD44780_puts("Hello!\n");

	// Initialize SPI for communication with the ADC (accelerometer) and gyrometer
	// Page 25 of L3G datasheet gives SPI protocol
	HD44780_putchar('0' + SPIIsRunning());
	if ((err = L3GStart(MOSI, MISO, SCLK, GYRO_CS, L3G_2000_DPS)))
		error(err, NULL);
	HD44780_putchar('0' + SPIIsRunning());
	if ((err = MCP300xStart(MOSI, MISO, SCLK, ADC_CS)))
		error(err, NULL);

	L3GAlwaysSetMode(1);
	MCP300xAlwaysSetMode(1);

	// Initialize filters
	for (i = 0; i < NUM_FILTERS; ++i) {
		filters[i].valueArray = filterValues[i];
		filters[i].weightArray = filterWeights[i];
		filters[i].head = filters[i].tail = 0;

		switch (i) {
			case ACCELEROMETER:
				filters[ACCELEROMETER].size = ACCEL_ARRAY_SIZE;
				for (j = 0; j < ACCEL_ARRAY_SIZE; ++j)
					filters[ACCELEROMETER].weightArray[j] =
					ACCEL_WEIGHT_FUNCTION;
				break;
			case GYROSCOPE:
				filters[GYROSCOPE].size = GYRO_ARRAY_SIZE;
				for (j = 0; j < ACCEL_ARRAY_SIZE; ++j)
					filters[GYROSCOPE].weightArray[j] = GYRO_WEIGHT_FUNCTION;
				break;
			default:
				error(FORGOT_TO_IMPLEMENT_FILTER, NULL);
		}
	}
}

static int8_t readEssentials (uint32_t *prevTimeCnt, filter_t filters[]) {
	int8_t err;
	static uint8_t first = 1; // Cleared after the first gyro calculation
	float dt = 0;
	uint32_t newTimeCnt = 0;
	int8_t accelVal;
	int16_t gyroY;

	uint16_t analogAccel;

	// Read values from Gyro and accelerometer
	if ((err = L3GReadY(&gyroY))) //pull Y-axis data from gyroscope
		error(err, NULL);
	if ((err = MCP300xRead(ACCEL, &analogAccel)))
		error(err, NULL);

	gyroY = runFilter(filters[GYROSCOPE], gyroY);
	analogAccel = runFilter(filters[ACCELEROMETER], analogAccel);

	// Calculate dt (time passed since previous read)
	newTimeCnt = CNT;
	if (!first)
		dt = ((float) (newTimeCnt - *prevTimeCnt)) / CLKFREQ;
	else
		first = 0;
	*prevTimeCnt = newTimeCnt;

	accelVal = (ACCEL_CENTER - analogAccel) / 3; // Correct math is: ... * 90degrees / 246steps, but that looked too hard

	g_leanAngle = ((1 - FILTER_TIME_CONSTANT)
			* (g_leanAngle + GYRO_WEIGHT * (gyroY * dt)))
			+ (FILTER_TIME_CONSTANT * accelVal); //value between -90 and 90 degrees

//	if (abs(g_leanAngle) > 80 && abs(accelVal) > 80)
//		//safety shutoff if the Segway falls over
//		error(FLAT_ON_FACE, NULL);

	return 0;
}

static int8_t readNonEssentials (void) {
	int8_t err;
	int16_t analogTurn;

	// Update turn angle once
	if ((err = MCP300xRead(TURN_POT, (uint16_t *) &analogTurn)))
		error(err, NULL);
	g_turnAngle = (TURN_CENTER - analogTurn) >> 2; // Correct math is: ... * 270degrees / 1024steps, but that looked even harder

	// Adjusts the turn angle to create a null zone
	if (g_turnAngle < -TURN_NULL_ZONE)
		g_turnAngle += TURN_NULL_ZONE;
	else if (g_turnAngle > TURN_NULL_ZONE)
		g_turnAngle -= TURN_NULL_ZONE;
	else
		g_turnAngle = 0;

	// Limit turn reading to within -MAX_TURN to MAX_TURN
	if (g_turnAngle > MAX_TURN_POWER)
		g_turnAngle = MAX_TURN_POWER;
	else if (g_turnAngle < -MAX_TURN_POWER)
		g_turnAngle = -MAX_TURN_POWER;

	return 0;
}

static int16_t runFilter (filter_t filter, int16_t newVal) {
	filter_size_t i, nextIndex;
	int32_t sum = 0;

	filter.valueArray[(filter.head)++] = newVal;

	nextIndex = filter.tail;
	for (i = 0; i < filter.size; ++i) {
		sum += filter.weightArray[i] * filter.valueArray[nextIndex];

		if (nextIndex == filter.size - 1)
			nextIndex = 0;
		else
			nextIndex++;

		if (nextIndex == filter.head)
			break;
	}

	sum /= 100;

	return sum;
}
